/*
  Test program for fl_read_image().
  
  The program creates a small image at the top left window position,
  reads this image with fl_read_image() with different widths and heights,
  and displays the result with a zoom factor of 4x at 4 different positions.
  
  The window can be resized, thus reading the image again, to see how it
  behaves (currently *very* slow on Windows with fl_read_image()).
  
  There is also a timer callback that resets and redraws the image at the
  lower right position every 5 seconds. This shows a potential bug (?),
  if make_current() is not used during the draw() method. The image is
  not drawn, or the space is white in my tests.
  
  This test file includes Roman Kantor's proposed version of
  fl_read_image_win32(). This doesn't work with normal windows, but may
  work with Fl_Offscreen's, due to the fact that fl_window is a
  *different* object type in FLTK's current implementation:
    (a) a Window (HWND) when reading directly from a window
    (b) a Bitmap (HBITMAP) when reading from an Fl_Offscreen
  This ought to be solvable somehow.

  There are some check buttons to modify the program's behavior:
  
   (1)	[rfw]	"read from window ?" (default: off)
		if this is off, then the original image data are displayed,
		for comparison of display times, otherwise fl_read_image()
		is used.
   (2)	[urv]	"use Roman's version (WIN only) ?" (default: off)
		you *must* also set [rfw] ON for this to work.
		The current behavior is undefined, may also crash !!!
		USE WITH CARE ! --- until we fix this ---
   (3)	[umc]	"use make_current() ?" (default: off)
		There seems to be a bug that the image is not drawn correctly,
		if it is drawn from the timer callback, whereas it is drawn
		correctly when the whole window is drawn after a resize.
		This needs investigation...

  Compile and run this program with:
  
  $ fltk-config --compile read_image.cxx && ./read_image

*/
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_RGB_Image.H>
#include <FL/Fl_Check_Button.H>
#include <FL/fl_draw.H>
#include <FL/x.H>

#include <stdio.h>

/*
  Change these definitions, if you want different image sizes.
  Note that IW/IC should be an integer (divisible w/o remainder).
*/
#define IW	(60)		// image width and height
#define IC	(10)		// # of image columns
#define XF	(IW/IC)		// column factor

/*
  Some global variables for easier testing...
*/

Fl_Window	*window = 0;
Fl_Check_Button *rfw = 0;	// read from window ?
Fl_Check_Button *urv = 0;	// use Roman's version: fl_read_image_win32()
Fl_Check_Button *umc = 0;	// use make_current()

uchar		*rgb_data = 0;	// original RGB pixel data for image

// =============================================================================

#ifdef	WIN32
#include <windows.h>

unsigned char * fl_read_image_win32( unsigned char *data, int x, int y, int w, int h, int alpha=0);

// -----------------------------------------------------------------
// this is Roman Kantor's code, modified with a few test statments:
// -----------------------------------------------------------------

unsigned char * fl_read_image_win32( unsigned char *data, int x, int y, int w, int h, int alpha){
  // RK:
  // The implementation reads the whole array with one call to GetDIBits()
  // (We could save some temporary memory if we would read the data on per-line basis instead but
  // it would be little bit slower - though still much fasrter than per-pixel readout)
  //
  // We also need to do some manimulation as - unlike in fltk - the DIB bits have format:
  // 1) lines are ordered from bottom to top
  // 2) the channels are in BGRA order, not RGB(A)


  if(!data){ // allocate memory for new data if buffer not provided
    if(alpha)
      data = new unsigned char[w * h * 4];
    else
      data = new unsigned char[w * h * 3];
  }

  BITMAP of_info;
  int suc = GetObject(fl_window, sizeof(BITMAP),&of_info);

  int err = GetLastError();
  printf ("GetObject returned: %d (%d,%d)\n",suc,of_info.bmWidth,of_info.bmHeight);
  printf ("GetLastError() = %d\n",err);
  fflush(stdout);

  BITMAPINFO   bi;
  bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  bi.bmiHeader.biWidth = window->w(); // of_info.bmWidth;	// changed for testing !!!
  bi.bmiHeader.biHeight = window->h(); // of_info.bmHeight;	// changed for testing !!!
  bi.bmiHeader.biPlanes = 1;
  bi.bmiHeader.biBitCount = 32;
  bi.bmiHeader.biCompression = BI_RGB;
  bi.bmiHeader.biSizeImage = 0;
  bi.bmiHeader.biXPelsPerMeter = 0;
  bi.bmiHeader.biYPelsPerMeter = 0;
  bi.bmiHeader.biClrUsed = 0;
  bi.bmiHeader.biClrImportant = 0;

  // Assure that we are not trying to read non-existing data.
  // If it is so, the function  should still work but the out-of-bounrady
  // part of the image is untouched (uninitialised if new array is allocated - so full of rubbish)

  int shift_x = 0; // X target shift if x modified
  int shift_y = 0; // Y target shift if y modified
  if(x < 0){
    shift_x = -x;
    w += x;
    x = 0;
  }
  if(y < 0){
    shift_y = -y;
    h += y;
    y = 0;
  }
  int ww = w; // We need the original width for output data line size before possible modification

  if(x + w > bi.bmiHeader.biWidth)
     w = bi.bmiHeader.biWidth - x;
  if(y + h > bi.bmiHeader.biHeight)
     h = bi.bmiHeader.biHeight - y;

  int line_size = ((of_info.bmWidth * bi.bmiHeader.biBitCount + 31) / 32) * 4; // Each line is four-byte aligned

  int dib_size =  line_size * h;
  unsigned char * dib = new unsigned char[dib_size]; // create temporary buffer to read DIB

  // The obtained lines are ordered from botom-to-top so we need to read corresponding top-bottom mirrored image
  int ncopied = GetDIBits(fl_gc, (HBITMAP)fl_window, of_info.bmHeight - y - h, h, dib, (BITMAPINFO *)&bi, DIB_RGB_COLORS);

  err = GetLastError();
  printf ("GetDIBits returned: ncopied = %d\n",ncopied);
  printf ("GetLastError() = %d\n",err);
  fflush(stdout);

  if(alpha){ // four-byre result with alpha
    unsigned char a = alpha;
    for(int j = 0; j<h; j++){
      const unsigned char * src = dib +  (h - j - 1) * line_size  + x * 4; // flipped source line
      unsigned char * tg = data + (j + shift_y) * 4 * ww + shift_x * 4; // target line
      for(int i = 0; i<w; i++){
        // The channels are in BGRA order, we need to swap them to fltk RGBA standard
        unsigned char b = *src++;
        unsigned char g = *src++;
        unsigned char r = *src; src += 2; // skip alpha source
        *tg++ = r;
        *tg++ = g;
        *tg++ = b;
        *tg++ = a;
      }
    }
  }else{ // three byte result without alpha channel
    for(int j = 0; j<h; j++){
      const unsigned char * src = dib + (h - j - 1)  * line_size + x * 4; // flipped source line
      unsigned char * tg = data + (j + shift_y) * 3 * ww + shift_x * 3; // target line
      for(int i = 0; i<w; i++){
        unsigned char b = *src++;
        unsigned char g = *src++;
        unsigned char r = *src; src += 2; // skip alpha source
        *tg++ = r;
        *tg++ = g;
        *tg++ = b;
      }
    }
  }
  delete[] dib; // delete DIB temporary buffer
  return data;
}

#endif	// WIN32

// =============================================================================

class MyBox : public Fl_Box {
  int ix_,iy_,iw_,ih_;		// captured image's coordinates
  Fl_Image *img_;		// cached image

  void draw();
  void resize (int X, int Y, int W, int H);
public:
  MyBox(int x, int y, int w, int h,
      int ix, int iy, int iw, int ih);
  void reset();
}; // class MyBox

void MyBox::draw() {		// draw method
  Fl_Box::draw();		// draw the box
  if (!img_) {			// image not (yet) chached
    uchar	*px;		// internal pixel array
    Fl_Image	*temp;		// internal image
    if (rfw->value()) {		// read from window
      px = (uchar *)malloc(3*iw_*ih_);

      // the following statement is needed *ONLY* if the box
      // is being drawn from the draw_timer callback, but not
      // if it is being drawn after a window resize.
      // Try with and w/o it !
      // This seems to be a bug on Windows only. (?)

      if (umc->value())		  // use make_current() ?
        window()->make_current(); // why is this needed ?

#if defined(WIN32)
      if (urv->value())		// use Roman's version
        fl_read_image_win32(	// Roman's version
		px,		// pixel array
		ix_,iy_,	// window (x,y)
		iw_,ih_);	// original image size
      else
#endif
        fl_read_image(		// use FLTK's version
		px,		// pixel array
		ix_,iy_,	// window (x,y)
		iw_,ih_);	// original image size
      temp = new Fl_RGB_Image(px,iw_,ih_,3);
    } else {
      temp = new Fl_RGB_Image(rgb_data,IW,IW,3);
    }
    img_ = temp->copy(w(),h());	// resize and store the image
    delete temp;
    if (rfw->value())		// if read from window
      free (px);		// free pixel array
    }				// image not yet chached
  img_->draw(x(),y());		// draw the image
} // draw()

void MyBox::resize (int X, int Y, int W, int H) {
  if (img_) {			// delete cached image
    delete img_;
    img_ = 0;
  }
  Fl_Box::resize(X,Y,W,H);
} // resize()

MyBox::MyBox(int x, int y, int w, int h,
      int ix, int iy, int iw, int ih)
    : Fl_Box(x,y,w,h) {
  ix_ = ix;
  iy_ = iy;
  iw_ = iw;
  ih_ = ih;
  img_ = 0;
  box(FL_NO_BOX);
  align(FL_ALIGN_CENTER | FL_ALIGN_INSIDE);
}

void MyBox::reset() {
  if (img_) delete img_;
  img_ = 0;
}

MyBox *box1 = 0;
MyBox *box2 = 0;
MyBox *box3 = 0;
MyBox *box4 = 0;

void draw_timer(void *) {
  static int n = 0;
  printf("draw_timer(%d)\n",++n); fflush(stdout);
  Fl::repeat_timeout (5.0,draw_timer);
  box4->reset();
  box4->redraw();
}

int main(int argc, char **argv) {

  int i,j,x,y,w,h;
  int r,g,b;
  Fl_RGB_Image *img;

  // create the image (rgb_image = global)

  rgb_data = (uchar*)malloc(IW*IW*3);		// rgb image, IWxIW px.
  uchar *ip = rgb_data;				// first byte
  for (i=0; i<IW; i++) {			// set colors ...
    for (j=0; j<IW; j++) {
      r = (j/XF+1) * 255 / IC;
      g = (i/XF+1) * 255 / IC;
      b = (((i+j)/2)/XF+1) * 255 / (2*IC);
      *ip++ = r<255 ? r : 255;
      *ip++ = g<255 ? g : 255;
      *ip++ = b<255 ? b : 255;
    }
  }

  img = new Fl_RGB_Image (rgb_data,IW,IW,3);		// create local image

  window = new Fl_Window(40+IW*9,40+IW*9);

  Fl_Box *box = new Fl_Box(10,10,IW,IW);
  box->box(FL_NO_BOX);
  box->image (img);
  box->align(FL_ALIGN_CENTER | FL_ALIGN_INSIDE);

  Fl_Box *resize_box = new Fl_Box(10+IW,10+IW,window->w()-10-IW,window->h()-10-IW);
  resize_box->box(FL_NO_BOX);

  rfw = new Fl_Check_Button(20+IW, 5,150,20,"read from window ?");
  rfw->value(0);
  rfw->visible_focus(0);

  urv = new Fl_Check_Button(20+IW,30,150,20,"use Roman's version (WIN only) ?");
  urv->value(0);
  urv->visible_focus(0);

  umc = new Fl_Check_Button(20+IW,55,150,20,"use make_current() ?");
  umc->value(0);
  umc->visible_focus(0);

  box1 = new MyBox(20+IW,  20+IW,  IW*4,    IW*4,    10,10,IW,  IW);
  box2 = new MyBox(30+IW*5,20+IW,  (IW-1)*4,(IW-1)*4,10,10,IW-1,IW-1);
  box3 = new MyBox(20+IW,  30+IW*5,(IW-2)*4,(IW-2)*4,10,10,IW-2,IW-2);
  box4 = new MyBox(30+IW*5,30+IW*5,(IW-3)*4,(IW-3)*4,10,10,IW-3,IW-3);

  window->resizable(resize_box);
  window->size_range(100,100);
  window->end();
  window->show(argc, argv);

  Fl::add_timeout(5.0,draw_timer);

  return Fl::run();
}
